unit MineSweeper;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  Buttons, StdCtrls;
const
  FieldSize = 16;
  FieldSizePlus1 = FieldSize + 1;
  CellDispSize = 16;
  clOrange = TColor($0080FF);
type
  CellStatus = (isMine, isMarked, isRevealed);
  CellStatusSet = set of CellStatus;
  MineCell = record
    status    : CellStatusSet;
    neighbour : byte
  end;
  MineField = array[0..FieldSizePlus1, 0..FieldSizePlus1] of MineCell;
  
  { TForm1 }

  TForm1 = class(TForm)
    QuitButton: TButton;
    ResetButton: TButton;
    TimeLabel: TLabel;
    RemainLabel: TLabel;
    MainPaintBox: TPaintBox;
    MarkToggleBox: TToggleBox;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure MainPaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MainPaintBoxPaint(Sender: TObject);
    procedure QuitButtonClick(Sender: TObject);
    procedure ResetButtonClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    { private declarations }
    mfld : MineField;
    mnum : integer;
    remain : integer;
    timepast : word;
    running : boolean;
    procedure InitVars;
    procedure InitMF;
    procedure RevealMF(var g : TCanvas);
    procedure NormalMFView(var g : TCanvas);
    procedure revealblock(x, y : integer);
  public
    { public declarations }
  end; 

var
  Form1: TForm1; 

implementation

uses AboutForm;

function fnstr(i : integer) : string;
var s : string;
begin
  str(i, s);
  fnstr := s
end;

{ TForm1 }
procedure TForm1.InitVars;
begin
  InitMF;
  RemainLabel.caption := fnstr(remain);
  timepast := 0;
  running := true;
  Timer1.enabled := true
end;

procedure TForm1.FormCreate(Sender: TObject);
var x, y : integer;
begin
  randomize;
  fillchar(mfld, sizeof(mfld), 0);
  for x := 1 to FieldSize do
    for y := 1 to FieldSize do
      include(mfld[x, y].status, isRevealed);
  running := true
end;

procedure TForm1.MainPaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if not Timer1.Enabled then Form2.Show;
  x := succ(x div CellDispSize);
  y := succ(y div CellDispSize);
  if (x in [1..FieldSize]) and (y in [1..FieldSize]) then begin
    with mfld[x, y] do
      if (Button = mbRight) or MarkToggleBox.checked then begin
        if isMarked in status then begin
          exclude(status, isMarked);
          dec(remain)
        end else begin
          include(status, isMarked);
          inc(remain)
        end
      end else
        if isMine in status then begin
          running := false;
          Timer1.Enabled := false;
          remain := mnum
        end else
          revealblock(x, y);
    MainPaintBox.refresh
  end;
  RemainLabel.caption := fnstr(remain)
end;

procedure TForm1.revealblock(x, y : integer);
begin
  with mfld[x, y] do
    if [isMine, isRevealed] * status = [] then begin
      include(status, isRevealed);
      if neighbour = 0 then begin
        revealblock(x-1, y-1);
        revealblock(x-1, y  );
        revealblock(x-1, y+1);
        revealblock(x  , y-1);
        revealblock(x  , y+1);
        revealblock(x+1, y-1);
        revealblock(x+1, y  );
        revealblock(x+1, y+1)
      end
    end
end;

procedure TForm1.NormalMFView(var g : TCanvas);
var x, y, xx, yy : integer;
begin
  with g do begin
    Pen.Color := clBlack;
    xx := 0;
    for x := 1 to FieldSize do begin
      yy := 0;
      for y := 1 to FieldSize do begin
        with mfld[x, y] do
          if isMarked in status then begin
            Brush.Color := clYellow;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize)
          end else if isRevealed in status then begin
            Brush.Color := clWhite;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize);
            if neighbour > 0 then TextOut(succ(xx), succ(yy), fnstr(neighbour))
          end;
        inc(yy, CellDispSize)
      end;
      inc(xx, CellDispSize)
    end
  end
end;

procedure TForm1.RevealMF(var g : TCanvas);
var x, y, xx, yy : integer;
begin
  with g do begin
    Brush.Color := clWhite;
    FillRect(0, 0, 256,256);
    Pen.Color := clBlack;
    Brush.Color := clRed;
    xx := 0;
    for x := 1 to FieldSize do begin
      xx := pred(x) * CellDispSize;
      yy := 0;
      for y := 1 to FieldSize do begin
        with mfld[x, y] do
          if isMine in status then begin
            if isMarked in status then Brush.Color := clOrange
                                  else Brush.Color := clRed;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize)
          end else if neighbour > 0 then
            TextOut(succ(xx), succ(yy), fnstr(neighbour));
        inc(yy, CellDispSize)
      end;
      inc(xx, CellDispSize)
    end
  end
end;

procedure TForm1.MainPaintBoxPaint(Sender: TObject);
var x, y, xx, yy : integer;
begin
  if running then NormalMFView(MainPaintBox.Canvas)
             else RevealMF(MainPaintBox.Canvas);
  with MainPaintBox.Canvas do begin
    Pen.Color := clBlack;
    for x := 0 to FieldSize do begin
      xx := x * CellDispSize;
      line(xx, 0, xx, 256)
    end;
    for y := 0 to FieldSize do begin
      yy := y * CellDispSize;
      line(0, yy, 256, yy)
    end
  end
end;

procedure TForm1.QuitButtonClick(Sender: TObject);
begin
  close
end;

procedure TForm1.ResetButtonClick(Sender: TObject);
begin
  InitVars;
  MainPaintBox.refresh
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  inc(timepast);
  TimeLabel.Caption := fnstr(timepast)
end;

procedure TForm1.InitMF;
var i, x, y : integer;
begin
  fillchar(mfld, sizeof(mfld), 0);
  mnum := 26 + random(10);
  remain := mnum;
  for i := 1 to mnum do begin
    repeat
      x := succ(random(FieldSize));
      y := succ(random(FieldSize))
    until not (isMine in mfld[x, y].status);
    include(mfld[x, y].status, isMine);
    inc(mfld[x-1, y-1].neighbour);
    inc(mfld[x-1, y  ].neighbour);
    inc(mfld[x-1, y+1].neighbour);
    inc(mfld[x  , y-1].neighbour);
    inc(mfld[x  , y+1].neighbour);
    inc(mfld[x+1, y-1].neighbour);
    inc(mfld[x+1, y  ].neighbour);
    inc(mfld[x+1, y+1].neighbour)
  end;
  for i := 0 to FieldSizePlus1 do begin     (* sentinels *)
    mfld[i, 0].neighbour := 255;
    mfld[i, FieldSizePlus1].neighbour := 255
  end;
  for i := 0 to FieldSizePlus1 do begin     (* sentinels *)
    mfld[0, i].neighbour := 255;
    mfld[FieldSizePlus1, i].neighbour := 255
  end
end;

initialization
  {$I minesweeper.lrs}

end.

